import uuid

import pytest
from unittest.mock import AsyncMock, MagicMock
from dawnet_client import WebSocketClient, DAWNetFilePath
from dawnet_client import ui_param


# Example method to register
async def example_method_one(a: int, b: float, c: str, d: DAWNetFilePath):
    pass


@pytest.mark.asyncio
async def test_register_method_one():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_one" in client.method_details
    assert client.method_details["example_method_one"]["params"] == [
        {"name": "a", "type": "int", "default_value": 0, "ui_component": None},
        {"name": "b", "type": "float", "default_value": 0.0, "ui_component": None},
        {"name": "c", "type": "str", "default_value": "", "ui_component": None},
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]

    assert client.master_token != None
    assert client.dawnet_token != None
    assert client.dawnet_token != client.master_token


@pytest.mark.asyncio
async def test_register_method_one_with_no_name_no_description():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_one" in client.method_details
    assert client.method_details["example_method_one"]["params"] == [
        {"name": "a", "type": "int", "default_value": 0, "ui_component": None},
        {"name": "b", "type": "float", "default_value": 0.0, "ui_component": None},
        {"name": "c", "type": "str", "default_value": "", "ui_component": None},
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]

    assert client.method_details["example_method_one"]["name"] == "Default Name"
    assert (
        client.method_details["example_method_one"]["description"]
        == "Default Description"
    )

    assert client.master_token != None
    assert client.dawnet_token != None
    assert client.dawnet_token != client.master_token


@pytest.mark.asyncio
async def test_register_method_one_with_description():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    client.set_name("My Special Method")
    client.set_description("My Special Method Description")

    # EXECUTE
    await client.register_method(example_method_one)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_one" in client.method_details
    assert client.method_details["example_method_one"]["params"] == [
        {"name": "a", "type": "int", "default_value": 0, "ui_component": None},
        {"name": "b", "type": "float", "default_value": 0.0, "ui_component": None},
        {"name": "c", "type": "str", "default_value": "", "ui_component": None},
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]

    assert client.method_details["example_method_one"]["name"] == "My Special Method"
    assert (
        client.method_details["example_method_one"]["description"]
        == "My Special Method Description"
    )
    assert client.master_token != None
    assert client.dawnet_token != None
    assert client.dawnet_token != client.master_token


async def example_method_one_defaults(
    a: int = 5, b: float = 2.2, c: str = "hello", d: DAWNetFilePath = None
):
    pass


@pytest.mark.asyncio
async def test_register_method_one_defaults():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_defaults)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_one_defaults" in client.method_details
    assert client.method_details["example_method_one_defaults"]["params"] == [
        {"name": "a", "type": "int", "default_value": 5, "ui_component": None},
        {"name": "b", "type": "float", "default_value": 2.2, "ui_component": None},
        {"name": "c", "type": "str", "default_value": "hello", "ui_component": None},
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]


async def example_method_one_partial_defaults(
    a: int, b: float, c: str = "hello", d: DAWNetFilePath = None
):
    pass


@pytest.mark.asyncio
async def test_method_one_partial_defaults():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_partial_defaults)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_one_partial_defaults" in client.method_details
    assert client.method_details["example_method_one_partial_defaults"]["params"] == [
        {"name": "a", "type": "int", "default_value": 0, "ui_component": None},
        {"name": "b", "type": "float", "default_value": 0.0, "ui_component": None},
        {"name": "c", "type": "str", "default_value": "hello", "ui_component": None},
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]


@ui_param("a", "DAWNetNumberSlider", min=0, max=10, step=1, default=5)
async def example_method_one_with_decorators(
    a: int, b: float, c: str = "hello", d: DAWNetFilePath = None
):
    pass


@pytest.mark.asyncio
async def test_register_method_one_with_decorators():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_with_decorators)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_one_with_decorators" in client.method_details
    assert client.method_details["example_method_one_with_decorators"]["params"] == [
        {
            "name": "a",
            "type": "int",
            "default_value": 5,
            "min": 0,
            "max": 10,
            "step": 1,
            "ui_component": "DAWNetNumberSlider",
        },
        {"name": "b", "type": "float", "default_value": 0.0, "ui_component": None},
        {"name": "c", "type": "str", "default_value": "hello", "ui_component": None},
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]


@ui_param("a", "DAWNetNumberSlider", min=0, max=10, step=1, default=5)
@ui_param("b", "DAWNetNumberSlider", min=0.0, max=10.0, step=0.5, default=7.7)
async def example_method_one_with_multiple_decorators(
    a: int, b: float = 7.7, c: str = "hello", d: DAWNetFilePath = None
):
    pass


@pytest.mark.asyncio
async def test_register_method_one_with_multiple_decorators():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_one_with_multiple_decorators)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_one_with_multiple_decorators" in client.method_details
    assert client.method_details["example_method_one_with_multiple_decorators"][
        "params"
    ] == [
        {
            "name": "a",
            "type": "int",
            "default_value": 5,
            "min": 0,
            "max": 10,
            "step": 1,
            "ui_component": "DAWNetNumberSlider",
        },
        {
            "name": "b",
            "type": "float",
            "default_value": 7.7,
            "min": 0.0,
            "max": 10.0,
            "step": 0.5,
            "ui_component": "DAWNetNumberSlider",
        },
        {"name": "c", "type": "str", "default_value": "hello", "ui_component": None},
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]


@ui_param("a", "unsupported_ui_component", min=0, max=10, step=1, default=5)
async def example_method_one_with_unsupported_decorators(
    a: int, b: float, c: str = "hello", d: DAWNetFilePath = None
):
    pass


@pytest.mark.asyncio
async def test_register_method_one_with_unsupported_decorators():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE and ASSERT
    with pytest.raises(ValueError) as exc_info:
        await client.register_method(example_method_one_with_unsupported_decorators)

    assert "Unsupported UI component" in str(exc_info.value)


@ui_param("a", "DAWNetNumberSlider", min=0, max=10, fakeparam="fake", step=1, default=5)
async def example_method_one_with_unsupported_param(
    a: int, b: float, c: str = "hello", d: DAWNetFilePath = None
):
    pass


@pytest.mark.asyncio
async def test_register_method_one_with_unsupported_param():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE and ASSERT
    with pytest.raises(ValueError) as exc_info:
        await client.register_method(example_method_one_with_unsupported_param)

    assert "Unsupported UI param" in str(exc_info.value)


@ui_param("a", "DAWNetNumberSlider", min=0, max=10, default=5)
async def example_method_one_missing_required_param(
    a: int, b: float, c: str = "hello", d: DAWNetFilePath = None
):
    pass


@pytest.mark.asyncio
async def test_register_method_one_missing_required_param():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE and ASSERT
    with pytest.raises(ValueError) as exc_info:
        await client.register_method(example_method_one_missing_required_param)

    assert "Missing required param(s)" in str(exc_info.value)


@ui_param("c", "DAWNetMultiChoice", options=["one", "two", "three"], default="two")
async def example_method_multichoice(a: int, b: float, c: str, d: DAWNetFilePath):
    pass


@pytest.mark.asyncio
async def test_register_method_example_method_multichoice():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_multichoice)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_multichoice" in client.method_details
    assert client.method_details["example_method_multichoice"]["params"] == [
        {"name": "a", "type": "int", "ui_component": None, "default_value": 0},
        {"name": "b", "type": "float", "ui_component": None, "default_value": 0.0},
        {
            "name": "c",
            "type": "str",
            "options": ["one", "two", "three"],
            "default_value": "two",
            "ui_component": "DAWNetMultiChoice",
        },
        {
            "name": "d",
            "type": "DAWNetFilePath",
            "default_value": None,
            "ui_component": None,
        },
    ]


async def example_method_with_bool_no_default(a: bool):
    pass


@pytest.mark.asyncio
async def test_register_method_example_method_with_bool_no_default():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_with_bool_no_default)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_with_bool_no_default" in client.method_details
    assert client.method_details["example_method_with_bool_no_default"]["params"] == [
        {"name": "a", "type": "bool", "ui_component": None, "default_value": False}
    ]


async def example_method_with_bool_default_True(a: bool = True):
    pass


@pytest.mark.asyncio
async def test_register_method_example_method_with_bool_default_True():
    # SETUP
    client = WebSocketClient("127.0.0.1", "1234")
    client.connect = AsyncMock()
    client.set_token(str(uuid.uuid4()))

    # EXECUTE
    await client.register_method(example_method_with_bool_default_True)

    # ASSERTS
    client.connect.assert_called_once()
    assert "example_method_with_bool_default_True" in client.method_details
    assert client.method_details["example_method_with_bool_default_True"]["params"] == [
        {"name": "a", "type": "bool", "ui_component": None, "default_value": True}
    ]



# decorators.py
def ui_param(name, ui_component, **kwargs):
    def decorator(func):
        if not hasattr(func, "_ui_params"):
            func._ui_params = {}
        func._ui_params[name] = {"ui_component": ui_component, **kwargs}
        return func

    return decorator



import logging
from enum import Enum
from typing import Any, Dict
from threading import Thread

from .config import SENTRY_API_KEY
import sentry_sdk


class DNTag(Enum):
    DNToken = "dn_token"
    DNSystemType = "dn_system"
    DNMsgType = "dn_msg_type"
    DNMsgStage = "dn_msg_stage"
    DNMsg = "dn_msg"


class DNSystemType(Enum):
    DN_PLUGIN = "DN_PLUGIN"
    DN_DISCOVERY_SERVER = "DN_DISCOVERY_SERVER"
    DN_API_SERVER = "DN_API_SERVER"
    DN_CLIENT = "DN_CLIENT"


class DNMsgType(Enum):
    DN_ERROR = "DN_ERROR"
    DN_EVENT = "DN_EVENT"


class DNMsgStage(Enum):
    SET_MSG_PENDING = "UPDATE_MSG_PENDING"
    REPLY_TO_MSG = "REPLY_TO_MSG"
    UPDATE_MSG_STATUS = "UPDATE_MSG_STATUS"
    ABORT_MSG = "ABORT_MSG"
    SAVE_CONTRACT = "SAVE_CONTRACT"
    UPLOAD_ASSET = "UPLOAD_ASSET"
    CLIENT_CONNECTION = "CLIENT_CONNECTION"
    CLIENT_REG_CONTRACT = "CLIENT_REG_CONTRACT"
    CLIENT_REG_METHOD = "CLIENT_REG_METHOD"
    CLIENT_RUN_METHOD = "CLIENT_RUN_METHOD"
    CLIENT_DOWNLOAD_ASSET = "CLIENT_DOWNLOAD_ASSET"
    CLIENT_UPLOAD_ASSET = "CLIENT_UPLOAD_ASSET"
    CLIENT_CONVERT_DOWNLOAD = "CLIENT_CONVERT_DOWNLOAD"
    CLIENT_CONVERT_UPLOAD = "CLIENT_CONVERT_UPLOAD"
    CLIENT_SEND_RESULTS_MSG = "CLIENT_SEND_RESULTS_MSG"
    WS_RECEIVE_MSG = "WS_RECEIVE_MSG"
    WS_REG_TOKEN = "WS_REG_TOKEN"
    WS_REG_CONTRACT = "WS_REG_CONTRACT"
    WS_SEND_RESULTS = "WS_SEND_RESULTS"
    WS_UN_REG_TOKEN = "WS_UN_REG_TOKEN"


def before_send(event, hint):
    # Check if the event has tags and the specific tag 'dn_token'
    if "tags" in event:
        dn_token = event["tags"].get("dn_token")
        if dn_token and dn_token.strip() != "":
            return event  # Send the event as it contains 'dn_token' and it's not empty
    # If 'dn_token' is not present or empty, do not send the event
    return None


def traces_sampler(sampling_context):
    # Access transaction context to check its name
    transaction_context = sampling_context.get("transaction_context", {})
    transaction_name = transaction_context.get("name")

    # Sample only if the transaction is named 'customer.event'
    if transaction_name == "customer.event":
        return 1  # Send the transaction to Sentry
    else:
        return 0  # Do not send the transaction


# Initialize Sentry
sentry_sdk.init(
    dsn=SENTRY_API_KEY,
    # Set traces_sample_rate to 0 to disable automatic performance monitoring
    # traces_sample_rate=1,
    traces_sampler=traces_sampler,
    # before_send=before_send
)


class SentryEventLogger:
    def __init__(self, service_name: str = "system-type-not-set") -> None:
        # Setup logging
        logging.basicConfig(level=logging.INFO)
        self.service_name = service_name
        self.logger = logging.getLogger(service_name)

    def log_event(self, dn_token: str, event_info: Dict[str, Any]) -> None:
        thread = Thread(
            target=self._handle_event,
            args=(dn_token, DNMsgType.DN_EVENT.value, event_info),
        )
        thread.start()

    def log_error(self, dn_token: str, event_info: Dict[str, Any]) -> None:
        thread = Thread(
            target=self._handle_event,
            args=(dn_token, DNMsgType.DN_ERROR.value, event_info),
        )
        thread.start()

    def _handle_event(
        self, dn_token: str, dn_msg_type: str, event_info: Dict[str, Any]
    ) -> None:
        self._process_event(event_info)
        return

    #         with start_transaction(op="task", name="customer.event") as transaction:
    #             # Set a custom UUID as a tag
    #             transaction.set_tag(DNTag.DNToken.value, str(dn_token))
    #             transaction.set_tag(DNTag.DNSystemType.value, self.service_name)
    #             transaction.set_tag(DNTag.DNMsgType.value, dn_msg_type)
    #
    #             for key, value in event_info.items():
    #                 transaction.set_tag(key, value)
    #
    #             self._process_event(event_info)
    #             transaction.set_data("event_info", event_info)
    #             transaction.sampled = True

    def _process_event(self, event_info: Dict[str, Any]) -> None:
        self.logger.info(f"Processing event: {event_info}")


# Example usage
# event_logger = SentryEventLogger()
# event_info = {"type": "purchase", "customer_id": "12345"}
# custom_uuid = "123e4567-e89b-12d3-a456-426614174000"  # Replace with your custom UUID
# event_logger.log_event(custom_uuid, event_info)



import shutil
import os
import wave
import aifc

from pydub.utils import mediainfo
from dawnet_client.utils import get_audio_length, process_audio_file

def clean_up_resampled_files(file_path):
    # Check if the directory exists
    if os.path.exists(file_path):
        # Remove the directory and all its contents
        shutil.rmtree(file_path)
        print(f"Directory '{file_path}' has been removed.")
    else:
        print(f"Directory '{file_path}' does not exist.")


def test_process_audio_file():
    clean_up_resampled_files(os.path.join(os.path.dirname(__file__), "assets", "resampled"))

    test_files_dir = os.path.join(os.path.dirname(__file__), "assets")
    output_formats = ['wav', 'aif', 'aiff', 'mp3', 'flac']
    sample_rates = [22050, 32000, 44100, 48000]
    bit_depths = [16, 24]  # Example bit depths
    channels_list = [1, 2]

    for file in os.listdir(test_files_dir):
        if file.endswith(('.wav', '.mp3', '.aif', '.flac', '.ogg')):
            file_path = os.path.join(test_files_dir, file)

            for format in output_formats:
                for sample_rate in sample_rates:
                    for bit_depth in bit_depths:
                        for channels in channels_list:
                            if format in ['wav', 'aif', 'aiff']:
                                print(f"\nTesting file: {file_path}")
                                print(f"Target Specs - Format: {format}, Sample Rate: {sample_rate}, Bit Depth: {bit_depth}, Channels: {channels}")

                                processed_file = process_audio_file(
                                    file_path,
                                    target_format=format,
                                    target_sample_rate=sample_rate,
                                    target_bit_depth=bit_depth,
                                    target_channels=channels
                                )

                                # Determine the format and get the specifications
                                file_extension = os.path.splitext(processed_file)[1].lower()
                                if file_extension in ['.wav', '.aif', '.aiff']:
                                    opener = wave.open if file_extension == '.wav' else aifc.open
                                    with opener(processed_file, 'rb') as f:
                                        actual_channels = f.getnchannels()
                                        actual_sample_rate = f.getframerate()
                                        actual_bit_depth = f.getsampwidth() * 8
                                elif file_extension in ['.mp3', '.flac']:
                                    info = mediainfo(processed_file)
                                    actual_channels = int(info['channels']) if file_extension != '.flac' else channels
                                    actual_sample_rate = int(info['sample_rate'])
                                    actual_bit_depth = int(info.get('bits_per_sample') or bit_depth) if file_extension != '.mp3' else bit_depth

                                print(f"Output Specs - File: {processed_file}, Sample Rate: {actual_sample_rate}, Bit Depth: {bit_depth}, Channels: {channels}")

                                # Assertions to ensure specifications match
                                assert actual_sample_rate == sample_rate, f"Sample rate mismatch for format {format} and sample rate {sample_rate}: expected {sample_rate}, got {actual_sample_rate}"
                                if file_extension not in ['.mp3', '.flac']:
                                    assert actual_bit_depth == bit_depth, f"Bit depth mismatch for format {format}: expected {bit_depth}, got {actual_bit_depth}"
                                    assert actual_channels == channels, f"Channel count mismatch for format {format}: expected {channels}, got {actual_channels}"

    clean_up_resampled_files(os.path.join(os.path.dirname(__file__), "assets", "resampled"))


def test_get_audio_length():
    test_files_dir = os.path.join(os.path.dirname(__file__), "assets")

    for file in os.listdir(test_files_dir):
        if file.endswith(('.wav', '.mp3', '.aif', '.flac', '.ogg')):
            file_path = os.path.join(test_files_dir, file)

            length_float = get_audio_length(file_path)

            rounded_length = round(length_float, 1)

            assert rounded_length == 16.0



import uuid

import pytest
from unittest.mock import patch, MagicMock
import dawnet_client


# Mock class to replace WebSocketClient
class MockWebSocketClient:
    def __init__(self, server_ip, server_port):
        self.input_format = None
        self.input_bit_depth = None
        self.input_sample_rate = None
        self.input_channels = None

        self.output_sample_rate = None
        self.output_bit_depth = None
        self.output_channels = None
        self.output_format = None

        self.dawnet_token = None

    # You might need to add mock methods here if they are called in your tests


# Patching the WebSocketClient class in dawnet_client.core module
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_format_valid():
    dawnet_client.set_input_target_format("mp3")
    assert dawnet_client.core._client.input_format == "mp3"


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_format_invalid():
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_input_target_format("abc")
    assert "Invalid format: 'abc'" in str(excinfo.value)


# Patching the WebSocketClient class in dawnet_client.core module
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_bit_depth_valid():
    # Test for valid bit depth 16
    dawnet_client.set_input_target_bit_depth(16)
    assert dawnet_client.core._client.input_bit_depth == 16

    # Test for valid bit depth 24
    dawnet_client.set_input_target_bit_depth(24)
    assert dawnet_client.core._client.input_bit_depth == 24


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_bit_depth_invalid():
    # Test for invalid bit depth
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_input_target_bit_depth(
            8
        )  # Assuming 8 is an invalid bit depth
    assert "Invalid bit depth: '8'" in str(excinfo.value)


# Patching the WebSocketClient class in dawnet_client.core module
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_sample_rate_valid():
    # Test for valid sample rates
    for rate in [22050, 32000, 44100, 48000]:
        dawnet_client.set_input_target_sample_rate(rate)
        assert dawnet_client.core._client.input_sample_rate == rate


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_sample_rate_invalid():
    # Test for invalid sample rate
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_input_target_sample_rate(
            16000
        )  # Assuming 16000 is an invalid sample rate
    assert "Invalid sample rate: '16000'" in str(excinfo.value)


# Patching the WebSocketClient class in dawnet_client.core module
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_channels_valid():
    # Test for valid channel counts
    for channels in [1, 2]:
        dawnet_client.set_input_target_channels(channels)
        assert dawnet_client.core._client.input_channels == channels


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_input_target_channels_invalid():
    # Test for invalid channel count
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_input_target_channels(
            3
        )  # Assuming 3 is an invalid channel count
    assert "Invalid channel count: '3'" in str(excinfo.value)


# Patching the WebSocketClient class in dawnet_client.core module
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_token_valid():
    # Generate a valid UUID4 token
    valid_uuid4 = str(uuid.uuid4())
    dawnet_client.set_token(valid_uuid4)
    assert dawnet_client.core._client.master_token == valid_uuid4


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_token_invalid():
    # Test with an invalid token
    invalid_token = "12345"
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_token(invalid_token)
    assert f"Invalid token: '{invalid_token}'" in str(excinfo.value)


# Test for set_output_target_sample_rate
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_sample_rate_valid():
    for rate in [22050, 32000, 44100, 48000]:
        dawnet_client.set_output_target_sample_rate(rate)
        assert dawnet_client.core._client.output_sample_rate == rate


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_sample_rate_invalid():
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_output_target_sample_rate(
            16000
        )  # Assuming 16000 is an invalid sample rate
    assert (
        "Invalid output sample rate: '16000'. Valid rates: [22050, 32000, 44100, 48000]"
        in str(excinfo.value)
    )


# Test for set_output_target_bit_depth
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_bit_depth_valid():
    for depth in [16, 24]:
        dawnet_client.set_output_target_bit_depth(depth)
        assert dawnet_client.core._client.output_bit_depth == depth


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_bit_depth_invalid():
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_output_target_bit_depth(8)
    assert "Invalid output bit depth: '8'. Valid depths: [16, 24]" in str(excinfo.value)


# Test for set_output_target_channels
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_channels_valid():
    for channels in [1, 2]:
        dawnet_client.set_output_target_channels(channels)
        assert dawnet_client.core._client.output_channels == channels


@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_channels_invalid():
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_output_target_channels(
            3
        )  # Assuming 3 is an invalid channel count
    assert "Invalid output channel count: '3'. Valid counts: [1, 2]" in str(
        excinfo.value
    )


# Test for set_output_target_format
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_format_valid():
    dawnet_client.set_output_target_format("mp3")
    assert dawnet_client.core._client.output_format == "mp3"


# Test for set_output_target_format
@patch("dawnet_client.core.WebSocketClient", new=MockWebSocketClient)
def test_set_output_target_format_invalid():
    with pytest.raises(ValueError) as excinfo:
        dawnet_client.set_output_target_format("abc")
    assert (
        "Invalid output format: 'abc'. Valid formats: ['wav', 'mp3', 'aif', 'aiff', 'flac']"
        in str(excinfo.value)
    )



import asyncio
import io
import sys
import uuid

import websockets
import nest_asyncio
import json
import logging
import os
import tempfile
import aiohttp

from .utils import process_audio_file
from .output import ResultsHandler
from .config import SOCKET_IP, SOCKET_PORT
from .dn_tracer import SentryEventLogger, DNSystemType, DNTag, DNMsgStage
from inspect import signature, Parameter

# Apply nest_asyncio to allow nested running of event loops
nest_asyncio.apply()

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)


class RunStatus:
    def __init__(self):
        self.status = "idle"


run_status = RunStatus()

# Method registry for the client
method_registry_local = {}
method_details_local = {}


class WebSocketClient:
    def __init__(self, server_ip, server_port):
        self.server_ip = server_ip
        self.server_port = server_port
        self.websocket = None
        self.method_registry = {}
        self.method_details = {}
        self.run_status = "idle"
        self.dawnet_token = None
        self.master_token = None
        self.message_id = None
        self.results = None
        self.author = "Default Author"
        self.name = "Default Name"
        self.description = "Default Description"
        self.version = "0.0.0"
        self.logger = logging.getLogger(__name__)
        self.dn_tracer = SentryEventLogger(service_name=DNSystemType.DN_CLIENT.value)

        # Default input target audio settings
        self.input_sample_rate = 44100
        self.input_bit_depth = 16
        self.input_channels = 2
        self.input_format = "wav"  # "wav", "mp3", "aif", "flac"

        # Default output target audio settings
        self.output_sample_rate = 44100
        self.output_bit_depth = 16
        self.output_channels = 2
        self.output_format = "wav"  # "wav", "mp3", "aif", "flac"

        # DAW SESSION INFO
        self.daw_bpm = 0
        self.daw_sample_rate = 0

    async def send_registered_methods_to_server(self):
        if self.dawnet_token is None:
            raise Exception(
                "Token not set. Please call set_token(token) before calling send_registered_methods_to_server()."
            )

        if self.master_token is None:
            raise Exception(
                "Master Token not set. Please call set_token(token) before calling send_registered_methods_to_server()."
            )

        await self.connect()  # Ensure we're connected

        # Check if there's at least one method registered
        if self.method_registry:
            # Get the last registered method's name and details
            last_method_name, last_method = next(reversed(self.method_registry.items()))
            last_method_details = self.method_details[last_method_name]

            # Construct the message to register the method
            register_compute_contract_msg = {
                "token": self.dawnet_token,
                "type": "contract",
                "data": last_method_details,
            }

            # Send the registration message to the server
            await self.websocket.send(json.dumps(register_compute_contract_msg))
            self.dn_tracer.log_event(
                self.dawnet_token,
                {
                    DNTag.DNMsgStage.value: DNMsgStage.CLIENT_REG_CONTRACT.value,
                    DNTag.DNMsg.value: f"Sent contract for registration. Token: {self.dawnet_token}",
                },
            )

    async def connect(self):
        if self.websocket is None or self.websocket.closed:
            uri = f"ws://{self.server_ip}:{self.server_port}"
            self.websocket = await websockets.connect(uri)
            self.dn_tracer.log_event(
                self.dawnet_token,
                {
                    DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONNECTION.value,
                    DNTag.DNMsg.value: f"Connected to {uri}",
                },
            )
            self.results = ResultsHandler(
                websocket=self.websocket,
                token=self.dawnet_token,
                target_sample_rate=self.output_sample_rate,
                target_bit_depth=self.output_bit_depth,
                target_channels=self.output_channels,
                target_format=self.output_format,
            )

        try:
            await self.register_compute_instance()
        except Exception as e:
            self.dn_tracer.log_error(
                self.dawnet_token,
                {
                    DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONNECTION.value,
                    DNTag.DNMsg.value: f"Error connecting. {e}",
                },
            )

    async def register_compute_instance(self):
        if self.dawnet_token is None:
            raise Exception(
                "Token not set. Please call set_token(token) before registering a method."
            )

        if self.master_token is None:
            raise Exception(
                "Master Token not set. Please call set_token(token) before registering a method."
            )

        # Construct the message to register the compute instance
        register_compute_instance_msg = {
            "token": self.dawnet_token,
            "type": "register",
            "data": {
                "master_token": self.master_token,
                "name": self.name,
                "description": self.description,
                "status": 1,  # Assuming 'status': 1 indicates a successful registration
            },
        }

        # Send the registration message to the server
        await self.websocket.send(json.dumps(register_compute_instance_msg))

    async def validate_and_process_parameters(self, method):
        params = []
        param_names = set()
        sig = signature(method)
        max_param_count = 12
        max_param_name_length = 36

        if len(sig.parameters) > max_param_count:
            raise ValueError("Method cannot have more than 12 parameters.")

        for param in sig.parameters.values():
            if len(param.name) > max_param_name_length:
                raise ValueError(
                    f"Parameter name '{param.name}' exceeds 36 characters."
                )

            if param.name in param_names:
                raise ValueError(f"Duplicate parameter name '{param.name}' detected.")
            param_names.add(param.name)

            if param.annotation is Parameter.empty:
                raise ValueError(
                    f"Parameter '{param.name}' is missing a type annotation."
                )

            param_type_name = param.annotation.__name__
            supported_types = {"bool", "int", "float", "str", "DAWNetFilePath"}
            if param_type_name not in supported_types:
                raise ValueError(
                    f"Unsupported type '{param_type_name}' for parameter '{param.name}'."
                )

            default_value = None if param.default is Parameter.empty else param.default
            default_value = await self.set_default_for_types(
                default_value, param_type_name
            )

            ui_component_details = self.process_ui_components(method, param)
            param_info = {
                "name": param.name,
                "type": param_type_name,
                "default_value": default_value,
            }
            param_info.update(ui_component_details)

            params.append(param_info)

        return params

    def process_ui_components(self, method, param):
        ui_component_details = {"ui_component": None}
        supported_ui_param_keys = {
            "min",
            "max",
            "step",
            "default",
            "ui_component",
            "options",
        }
        supported_ui_components = {"DAWNetNumberSlider", "DAWNetMultiChoice"}
        ui_component_requirements = {
            "dawnetnumberslider": {"min", "max", "step", "default"},
            "dawnetmultichoice": {"options", "default"},
        }

        if hasattr(method, "_ui_params") and param.name in method._ui_params:
            ui_param_info = method._ui_params[param.name]

            for key in ui_param_info.keys():
                if key not in supported_ui_param_keys:
                    raise ValueError(
                        f"Unsupported UI param '{key}' for parameter '{param.name}'."
                    )

            if "ui_component" in ui_param_info:
                ui_component = ui_param_info["ui_component"].lower()
                required_params = ui_component_requirements.get(ui_component, set())
                missing_params = required_params - set(
                    key.lower() for key in ui_param_info.keys()
                )
                if missing_params:
                    raise ValueError(
                        f"Missing required param(s) {missing_params} for UI component '{ui_component}' in parameter '{param.name}'."
                    )

                if ui_component not in {
                    comp.lower() for comp in supported_ui_components
                }:
                    raise ValueError(
                        f"Unsupported UI component '{ui_component}' for parameter '{param.name}'."
                    )

                ui_component_details["ui_component"] = ui_component
                ui_component_details.update(
                    {k: v for k, v in ui_param_info.items() if k != "default"}
                )

            if "default" in ui_param_info:
                ui_component_details["default_value"] = ui_param_info["default"]

        return ui_component_details

    def create_json_payload(self, method_name, params):
        method_details = {
            "method_name": method_name,
            "params": params,
            "author": self.author,
            "name": self.name,
            "description": self.description,
            "version": self.version,
        }
        return method_details

    def generate_uuid(self, method_details):
        method_details_str = json.dumps(method_details, sort_keys=True)
        return str(uuid.uuid5(uuid.UUID(self.master_token), method_details_str))

    async def set_default_for_types(self, default_value, param_type_name):
        # if the type is bool and the default value is None, set it to False
        if param_type_name == "bool" and default_value is None:
            default_value = False
        if param_type_name == "int" and default_value is None:
            default_value = 0
        if param_type_name == "float" and default_value is None:
            default_value = 0.0
        if param_type_name == "str" and default_value is None:
            default_value = ""
        return default_value

    async def register_method(self, method):
        if self.master_token is None:
            raise Exception(
                "Master Token not set. Please call set_token(token) before registering a method."
            )

        if not asyncio.iscoroutinefunction(method):
            raise ValueError("The method must be asynchronous (async).")

        method_name = method.__name__
        params = await self.validate_and_process_parameters(method)
        method_details = self.create_json_payload(method_name, params)
        self.method_details[method_name] = method_details
        self.method_registry = {method_name: method}

        #TODO: there should be a better way to do this:
        self.dawnet_token = self.generate_uuid(
            str(method_details) + str(self.master_token) + str(self.name) + str(self.version) + str(self.author) + str(
                self.description))

        await self.connect()  # Ensure we're connected

        self.dn_tracer.log_event(
            self.dawnet_token,
            {
                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_REG_METHOD.value,
                DNTag.DNMsg.value: f"Registered method: {method_name}",
            },
        )

    async def run_method(self, name, **kwargs):
        run_status.status = "running"
        if name in self.method_registry:
            method = self.method_registry[name]
            if asyncio.iscoroutinefunction(method):
                print("IS COROUTINE")
                # If the method is a coroutine, await it directly
                try:
                    # Capture stdout and stderr
                    stdout_buffer = io.StringIO()
                    stderr_buffer = io.StringIO()
                    sys.stdout = stdout_buffer
                    sys.stderr = stderr_buffer

                    await method(**kwargs)

                    # Restore the original stdout and stderr
                    sys.stdout = sys.__stdout__
                    sys.stderr = sys.__stderr__

                    # Get the captured output as strings
                    stdout_output = stdout_buffer.getvalue()
                    stderr_output = stderr_buffer.getvalue()

                    # Now you can do whatever you want with the output
                    print("Captured stdout:")
                    print(stdout_output)
                    print("Captured stderr:")
                    print(stderr_output)

                    await self.results.add_log(stdout_output)
                    await self.results.add_log(stderr_output)

                    # If you want to save the output to a file, you can do so here
                    # with open("stdout.log", "w") as stdout_file:
                    #     stdout_file.write(stdout_output)
                    #
                    # with open("stderr.log", "w") as stderr_file:
                    #     stderr_file.write(stderr_output)

                    # if self.results.errors is None:
                    #     self.results.add_error("ENCOUNTERED AN ERROR")

                    print("ERRORS: " + str(self.results.errors))

                    await self.results.send()

                    self.dn_tracer.log_event(
                        self.dawnet_token,
                        {
                            DNTag.DNMsgStage.value: DNMsgStage.CLIENT_RUN_METHOD.value,
                            DNTag.DNMsg.value: f"Ran method: {name}",
                        },
                    )
                except Exception as e:
                    await self.results.add_error("ERROR:" + str(e))
                    print("IM IN THE EXCEPTION")
                    await self.results.send()

                    self.dn_tracer.log_error(
                        self.dawnet_token,
                        {
                            DNTag.DNMsgStage.value: DNMsgStage.CLIENT_RUN_METHOD.value,
                            DNTag.DNMsg.value: f"Error running method: {e}",
                        },
                    )
            else:
                print("IS NOT COROUTINE")
                # If the method is not a coroutine, run it in an executor
                # loop = asyncio.get_running_loop()
                # try:
                #     func = lambda: method(**kwargs)
                #     result = await loop.run_in_executor(None, func)
                #     self.dn_tracer.log_event(self.dawnet_token, {
                #         DNTag.DNMsgStage.value: DNMsgStage.CLIENT_RUN_METHOD.value,
                #         DNTag.DNMsg.value: f"Ran method: {name}",
                #     })
                # except Exception as e:
                #     self.dn_tracer.log_error(self.dawnet_token, {
                #         DNTag.DNMsgStage.value: DNMsgStage.CLIENT_RUN_METHOD.value,
                #         DNTag.DNMsg.value: f"Error running method: {e}",
                #     })

            run_status.status = "stopped"
            return True
        else:
            run_status.status = "stopped"
            raise Exception("Method not registered")

    async def download_gcp_files(self, obj, session):
        """
        Recursively search for GCP URLs in a JSON object and download the files.
        """
        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, str) and value.startswith(
                        "https://storage.googleapis.com"
                ):
                    # Download and replace the URL with a local file path
                    try:
                        obj[key] = await self.download_file(value, session)

                        self.dn_tracer.log_event(
                            self.dawnet_token,
                            {
                                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_DOWNLOAD_ASSET.value,
                                DNTag.DNMsg.value: f"Downloaded: {str(obj[key])}",
                            },
                        )
                    except Exception as e:
                        self.dn_tracer.log_error(
                            self.dawnet_token,
                            {
                                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_DOWNLOAD_ASSET.value,
                                DNTag.DNMsg.value: f"Error downloading: {e}",
                            },
                        )
                elif isinstance(value, (dict, list)):
                    await self.download_gcp_files(value, session)
        elif isinstance(obj, list):
            for item in obj:
                await self.download_gcp_files(item, session)

    async def download_file(self, url, session):
        """
        Download a file from a URL, save it to a temporary directory, and process if it's an audio file.
        """
        local_filename = url.split("/")[-1]
        local_path = os.path.join(self.temp_dir, local_filename)

        async with session.get(url) as response:
            if response.status == 200:
                with open(local_path, "wb") as f:
                    f.write(await response.read())

                # Check if the file is an audio file
                if os.path.splitext(local_path)[1][1:] in [
                    "wav",
                    "mp3",
                    "aif",
                    "aiff",
                    "flac",
                    "ogg",
                ]:
                    try:
                        local_path = process_audio_file(
                            local_path,
                            self.input_format,
                            self.input_sample_rate,
                            self.input_bit_depth,
                            self.input_channels,
                        )

                        self.dn_tracer.log_event(
                            self.dawnet_token,
                            {
                                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONVERT_DOWNLOAD.value,
                                DNTag.DNMsg.value: f"Converted download: {local_path}",
                            },
                        )
                    except Exception as e:
                        self.dn_tracer.log_error(
                            self.dawnet_token,
                            {
                                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONVERT_DOWNLOAD.value,
                                DNTag.DNMsg.value: f"Error converting downloading: {e}",
                            },
                        )

                return local_path
            else:
                raise Exception(f"Failed to download file: {url}")

    async def listen(self):
        if self.dawnet_token is None:
            raise Exception(
                "Token not set. Please call set_token(token) before starting to listen."
            )

        await self.connect()  # Ensure we're connected

        try:
            # Create a temporary directory
            self.temp_dir = tempfile.mkdtemp()
            self.logger.info(f"Created a temporary directory: {self.temp_dir}")

            async with aiohttp.ClientSession() as session:
                # Continuous listening loop
                while True:
                    register_compute_instance_msg = await self.websocket.recv()

                    msg = json.loads(register_compute_instance_msg)

                    # Download GCP-hosted files and update the JSON
                    try:
                        await self.download_gcp_files(msg, session)
                    except Exception as e:
                        self.dn_tracer.log_error(
                            _client.dawnet_token,
                            {
                                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_DOWNLOAD_ASSET.value,
                                DNTag.DNMsg.value: f"Error downloading GCP files: {e}",
                            },
                        )

                    if 'type' in msg:
                        if msg["type"] == "run_method":
                            # Check if the status is already "running"
                            if run_status.status == "running":
                                await self.websocket.send("Plugin already started!")
                            else:
                                self.results.clear_outputs()  # Clear previous outputs before running the method
                                self.message_id = msg["message_id"]
                                self.results.set_message_id(self.message_id)
                                self.daw_bpm = msg["bpm"]
                                self.daw_sample_rate = msg["sample_rate"]

                                data = msg["data"]
                                method_name = data["method_name"]
                                # Extract 'value' for each parameter to build kwargs
                                params = {
                                    param_name: param_details["value"]
                                    for param_name, param_details in data["params"].items()
                                }

                                # Now you can call run_method using argument unpacking
                                asyncio.create_task(self.run_method(method_name, **params))
                        elif msg['type'] == 'close_connection':
                            try:
                                await self.websocket.close()
                            except Exception as e:
                                print("Error closing connection: ", e)

                            print("Connection closed by server")
                            break  # Exit the while loop

                    else:
                        self.dn_tracer.log_error(
                            _client.dawnet_token,
                            {
                                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONNECTION.value,
                                DNTag.DNMsg.value: "UNKNOWN MESSAGE TYPE",
                            },
                        )

        except websockets.exceptions.ConnectionClosedOK:
            self.dn_tracer.log_error(
                _client.dawnet_token,
                {
                    DNTag.DNMsgStage.value: DNMsgStage.CLIENT_CONNECTION.value,
                    DNTag.DNMsg.value: "Connection was closed.",
                },
            )

    def set_token(self, token):
        self.master_token = token
        # if self.results is not None:
        #     self.results.update_token(token)

    def set_author(self, author):
        self.author = author
        self.update_all_method_details()

    def set_name(self, name):
        self.name = name
        self.update_all_method_details()

    def set_description(self, description):
        self.description = description
        self.update_all_method_details()

    def set_version(self, version):
        self.version = version
        self.update_all_method_details()

    def update_all_method_details(self):
        for method_name, method_detail in self.method_details.items():
            # If method_detail is already a dict, no need to load it
            method_detail["author"] = self.author
            method_detail["name"] = self.name
            method_detail["description"] = self.description
            method_detail["version"] = self.version
            self.method_details[method_name] = method_detail  # If you need it as a dict
            # If you need to store it as a JSON string, then use json.dumps
            # self.method_details[method_name] = json.dumps(method_detail)

    def run(self):
        asyncio.run(self.listen())


# Create a single WebSocketClient instance
_client = WebSocketClient(SOCKET_IP, SOCKET_PORT)


def output():
    return _client.results


# Define the functions that will interact with the WebSocketClient instance
def set_token(token):
    try:
        # Create a UUID object from the token
        uuid_obj = uuid.UUID(token)

        # Check if the formatted string of the UUID object matches the input token
        # This comparison is case-insensitive and ignores hyphens
        if uuid_obj.hex != token.replace("-", "").lower():
            raise ValueError
    except ValueError:
        raise ValueError(f"Invalid token: '{token}'. Token must be a valid UUID.")

    _client.set_token(token)


async def _register_method(method):
    await _client.register_method(method)


def register_method(method):
    try:
        asyncio.run(_register_method(method))
    except Exception as e:
        dn_tracer = SentryEventLogger(service_name=DNSystemType.DN_CLIENT.value)
        dn_tracer.log_error(
            _client.dawnet_token,
            {
                DNTag.DNMsgStage.value: DNMsgStage.CLIENT_REG_METHOD.value,
                DNTag.DNMsg.value: f"Error registering method: {e}",
            },
        )


def set_author(author: str):
    _client.set_author(author)


def set_name(name: str):
    _client.set_name(name)


def set_description(description: str):
    _client.set_description(description)


def set_version(version):
    _client.set_version(version)


def set_input_target_sample_rate(sample_rate: int):
    # List of valid sample rates
    valid_sample_rates = [22050, 32000, 44100, 48000]

    # Check if the sample rate is valid
    if sample_rate in valid_sample_rates:
        _client.input_sample_rate = sample_rate
    else:
        # Raise an error if the sample rate is not valid
        raise ValueError(
            f"Invalid sample rate: '{sample_rate}'. Valid sample rates are: {', '.join(map(str, valid_sample_rates))}"
        )


def set_input_target_bit_depth(bit_depth: int):
    # List of valid bit depths
    valid_bit_depths = [16, 24]

    # Check if the bit depth is valid
    if bit_depth in valid_bit_depths:
        _client.input_bit_depth = bit_depth
    else:
        # Raise an error if the bit depth is not valid
        raise ValueError(
            f"Invalid bit depth: '{bit_depth}'. Valid bit depths are: {', '.join(map(str, valid_bit_depths))}"
        )


def set_input_target_channels(channels: int):
    # List of valid channel counts
    valid_channels = [1, 2]

    # Check if the channel count is valid
    if channels in valid_channels:
        _client.input_channels = channels
    else:
        # Raise an error if the channel count is not valid
        raise ValueError(
            f"Invalid channel count: '{channels}'. Valid channel counts are: {', '.join(map(str, valid_channels))}"
        )


def set_input_target_format(format: str):
    # List of valid formats (in lower case)
    valid_formats = ["wav", "mp3", "aif", "aiff", "flac"]

    # Convert the input format to lower case
    format_lower = format.lower()

    # Check if the format is in the list of valid formats
    if format_lower in valid_formats:
        _client.input_format = format_lower
    else:
        # Raise an error if the format is not valid
        raise ValueError(
            f"Invalid format: '{format}'. Valid formats are: {', '.join(valid_formats)}"
        )


def set_output_target_sample_rate(sample_rate: int):
    # Assuming the same valid sample rates as for input
    valid_sample_rates = [22050, 32000, 44100, 48000]
    if sample_rate in valid_sample_rates:
        _client.output_sample_rate = sample_rate
    else:
        raise ValueError(
            f"Invalid output sample rate: '{sample_rate}'. Valid rates: {valid_sample_rates}"
        )


def set_output_target_bit_depth(bit_depth: int):
    # Assuming the same valid bit depths as for input
    valid_bit_depths = [16, 24]
    if bit_depth in valid_bit_depths:
        _client.output_bit_depth = bit_depth
    else:
        raise ValueError(
            f"Invalid output bit depth: '{bit_depth}'. Valid depths: {valid_bit_depths}"
        )


def set_output_target_channels(channels: int):
    # Assuming the same valid channel counts as for input
    valid_channels = [1, 2]
    if channels in valid_channels:
        _client.output_channels = channels
    else:
        raise ValueError(
            f"Invalid output channel count: '{channels}'. Valid counts: {valid_channels}"
        )


def set_output_target_format(format: str):
    # Assuming the same valid formats as for input
    valid_formats = ["wav", "mp3", "aif", "aiff", "flac"]
    format_lower = format.lower()
    if format_lower in valid_formats:
        _client.output_format = format
    else:
        raise ValueError(
            f"Invalid output format: '{format}'. Valid formats: {valid_formats}"
        )


def get_daw_bpm():
    return _client.daw_bpm


def get_daw_sample_rate():
    return _client.daw_sample_rate


def connect_to_server():
    asyncio.run(_client.send_registered_methods_to_server())
    asyncio.run(_client.listen())


# THIS IS A SPECIAL TYPE THAT WILL BE USED TO REPRESENT FILE UPLOADS
class DAWNetFilePath(str):
    pass



import os
import librosa
import numpy as np
import soundfile as sf
from pydub import AudioSegment


def process_audio_file(
    file_path: str,
    target_format: str = "wav",
    target_sample_rate: int = 44100,
    target_bit_depth: int = 16,
    target_channels: int = 2,
):
    # Create 'resampled' directory if it doesn't exist
    resampled_dir = os.path.join(os.path.dirname(file_path), "resampled")
    if not os.path.exists(resampled_dir):
        os.makedirs(resampled_dir)

    # Set the output file extension and format
    base_name = os.path.splitext(os.path.basename(file_path))[0]
    output_file_extension = target_format.lower()
    output_file_path = os.path.join(
        resampled_dir, f"{base_name}.{output_file_extension}"
    )
    output_format = (
        "AIFF" if output_file_extension in ["aif", "aiff"] else output_file_extension
    )

    # Set the subtype for bit depth (specific for WAV and AIFF)
    subtype = "PCM_16"  # default
    if output_file_extension in ["wav", "aif", "aiff"]:
        subtype = f"PCM_{target_bit_depth}" if target_bit_depth in [16, 24] else subtype

    # Load and process the audio file
    y, sr = librosa.load(
        file_path, sr=None, mono=False
    )  # Load with original sample rate and preserve channels
    y = librosa.resample(y, orig_sr=sr, target_sr=target_sample_rate)  # Resample

    # Handle stereo or mono conversion
    if target_channels == 1:
        y = librosa.to_mono(y)
    elif target_channels == 2 and y.ndim == 1:
        y = np.vstack([y, y])

    # Convert to target bit depth
    if target_bit_depth == 16:
        y = (y * np.iinfo(np.int16).max).astype(np.int16)
    elif target_bit_depth == 24:
        # Process as 32-bit but write as 24-bit
        y = (y * np.iinfo(np.int32).max).astype(np.int32)

    # Write the audio file
    sf.write(
        output_file_path, y.T, target_sample_rate, format=output_format, subtype=subtype
    )

    # Verify the result
    # processed_audio = AudioSegment.from_file(output_file_path)
    # assert processed_audio.frame_rate == target_sample_rate, f"Sample rate conversion failed: expected {target_sample_rate}, got {processed_audio.frame_rate}"
    # assert processed_audio.channels == target_channels, f"Channel conversion failed: expected {target_channels}, got {processed_audio.channels}"
    # assert processed_audio.sample_width * 8 in [target_bit_depth, 32], f"Bit depth conversion failed: expected {target_bit_depth} or 32, got {processed_audio.sample_width * 8}"

    return output_file_path


def get_audio_length(file_path: str) -> float:
    """
    Returns the length of the audio file in seconds.

    :param file_path: Path to the audio file
    :return: Length of the audio in seconds
    """
    try:
        # Load the audio file
        audio = AudioSegment.from_file(file_path)

        # Calculate the length in seconds
        length_seconds = len(audio) / 1000.0
        return length_seconds
    except Exception as e:
        print(f"Error processing file {file_path}: {e}")
        return None



import os

from aiohttp import ClientSession
from .config import API_BASE_URL, STORAGE_BUCKET_PATH


class FileUploader:
    async def get_signed_url(self, filename, token) -> str:
        url = (
            f"{API_BASE_URL}/api/hub/get_signed_url/?token={token}&filename={filename}"
        )
        async with ClientSession() as session:
            async with session.get(url) as response:
                # TODO: Check if response is ok
                data = await response.json()
                return data["signed_url"]

    async def upload_file_to_gcp(self, file_path, signed_url, file_type) -> bool:
        async with ClientSession() as session:
            with open(file_path, "rb") as file:
                async with session.put(
                    signed_url, data=file, headers={"Content-Type": file_type}
                ) as response:
                    return response.status == 200

    async def upload(self, file_path, file_type) -> str:
        file_name = os.path.basename(file_path)
        storage_bucket_path = STORAGE_BUCKET_PATH.rstrip('/')
        file_url = f"{storage_bucket_path}/{file_name}"
        signed_url = await self.get_signed_url(file_name, "myToken")
        result = await self.upload_file_to_gcp(file_path, signed_url, file_type)

        if result:
            return file_url
        else:
            raise Exception("Failed to upload file to GCP")



from .core import connect_to_server, register_method, set_token, set_author, set_name, set_description, \
    set_version, set_input_target_format, set_input_target_channels, set_input_target_bit_depth, \
    set_input_target_sample_rate, DAWNetFilePath, set_output_target_channels, set_output_target_format, \
    set_output_target_bit_depth, set_output_target_sample_rate, output, WebSocketClient
from .dn_tracer import SentryEventLogger, DNSystemType, DNMsgType, DNMsgStage, DNTag
from . import utils
from . import output
from .decorators import ui_param



from setuptools import setup, find_packages
import subprocess


def is_ffmpeg_installed():
    try:
        subprocess.run(
            ["ffmpeg", "-version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        return True
    except FileNotFoundError:
        return False


ffmpeg_warning_msg = """
Warning: FFmpeg is not installed on your system, which is required for some functionalities of this package.
Please install FFmpeg using the following instructions:

- For macOS:
  Install FFmpeg using Homebrew: `brew install ffmpeg`

- For Linux (Debian/Ubuntu):
  Install FFmpeg using apt: `sudo apt-get install ffmpeg`

After installing, ensure that FFmpeg is in your PATH.
"""

if not is_ffmpeg_installed():
    print(ffmpeg_warning_msg)

setup(
    name="dawnet-client",
    version="0.6.2",
    packages=find_packages(),
    install_requires=[
        "websockets",
        "nest-asyncio",
        "sentry-sdk",
        "pydub",
        "librosa",
        "pytest-asyncio",
    ],
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "dawnet-client=dawnet_client.core:main",
        ],
    },
    # Additional metadata about your package
    author="Steve Hiehn",
    author_email="stevehiehn@gmail.com",
    description="DAWNet client enables remote execution of python code triggered from a DAW.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://dawnet.tools",
)



import os

# --------- PRODUCTION SETTINGS ----------------
SOCKET_IP = os.getenv("DN_CLIENT_SOCKET_IP", "34.135.228.111")
API_BASE_URL = os.getenv("DN_CLIENT_API_BASE_URL", "http://34.135.228.111:8081")
SENTRY_API_KEY = os.getenv(
    "DN_CLIENT_SENTRY_API_KEY",
    "https://dbbf2855a707e0448957ad77111449c3@o4506379662131200.ingest.sentry.io/4506379670847488",
)
SOCKET_PORT = os.getenv("DN_CLIENT_SOCKET_PORT", "8765")
STORAGE_BUCKET_PATH = os.getenv("DN_CLIENT_STORAGE_BUCKET", "https://storage.googleapis.com/byoc-file-transfer/")
